import functools
import inspect
import logging
from typing import (
    Any,
    Callable,
)

from connector.errors import (
    ConnectorError,
    DefaultHandler,
    ErrorMap,
)
from connector.serializers.lumos import (
    EncounteredErrorResponse,
    ErrorResp,
)
from connector.sync_.lumos import LumosCommandsMixin

logger = logging.getLogger(__name__)


def connector_handler(
    exception_classes: ErrorMap,
):
    """
    Decorator that adds default exception handlers to a class. Iterates over its methods and attaches a exception_handler decorator (below).

    Example:
    ```python
    @connector_handler([
        (httpx.HTTPStatusError, http_status_error, "error.code"),
    ])
    class AsyncCommands(LumosCommandsMixin):
        def verify_credentials(self, args: ValidateCredentialsArgs) -> ValidateCredentialsResp:
            pass
    ```

    Args:
        exception_classes (tuple): Tuple of exception classes to be handled. Map of exception class to handler function.
    Returns:
        class_decorator (function): Decorator that adds exception handlers to a class.
    """
    default_classes: ErrorMap = []
    exception_classes.extend(default_classes)

    def class_decorator(cls: LumosCommandsMixin):
        app_id = None
        attrs = inspect.getmembers(cls, lambda a: not inspect.isroutine(a))
        for name, value in attrs:
            if name == "app_id":
                app_id = value

        for attr_name, attr_value in cls.__dict__.items():
            if callable(attr_value):
                decorated_func = exception_handler(exception_classes, app_id)(attr_value)
                setattr(cls, attr_name, decorated_func)
        return cls

    return class_decorator


def exception_handler(
    exception_classes: ErrorMap,
    app_id: str | None = None,
):
    """
    Decorator that adds error handling to a method. Uses the default Lumos error handler if no exception handler is provided.

    Example:
    ```python
    @exception_handler(
        (httpx.HTTPStatusError, http_status_error, "error.code"),
    )
    async def verify_credentials(self, args: ValidateCredentialsArgs) -> ValidateCredentialsResp:
        pass
    ```

    Args:
        exception_classes (tuple): Tuple of exception classes to be handled. Map of exception class to handler function.
    Returns:
        function: Decorated function.
    """

    def exception_handler_decorator(func: Callable[[Any], Any]):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                resp = DefaultHandler.handle(
                    e,
                    func,
                    ErrorResp(response=EncounteredErrorResponse(message=str(e))),
                    app_id if app_id else None,
                )

                if not isinstance(e, ConnectorError):
                    for exception_class, handler, code in exception_classes:
                        if isinstance(e, exception_class) and handler:
                            resp.response.error_code = f"{app_id}.{code}" if app_id else code
                            resp = handler(e, func, resp, code)

                logger.error(f"{resp.response.error_code}: {resp.response.message}")
                return resp

        return wrapper

    return exception_handler_decorator



from connector.helpers import collect_methods


def collect_routes(obj: object):
    """
    Collect all methods from an object and create a route for each.
    """
    from fastapi import APIRouter

    router = APIRouter()
    commands = collect_methods(obj)
    for method in commands:
        router.add_api_route(f"/{method.__name__.replace('_', '-')}", method, methods=["POST"])
    return router


def runserver(router, port: int):
    import uvicorn
    from fastapi import FastAPI

    app = FastAPI()
    app.include_router(router)

    uvicorn.run(app, port=port)



from abc import abstractmethod
from enum import StrEnum
from types import FunctionType
from typing import Any, Callable, List, Tuple

from connector.serializers.lumos import ErrorResp


class ErrorCodes(StrEnum):
    """
    Error codes for Lumos connector.
    """

    NOT_FOUND = "not_found"
    INTERNAL_ERROR = "internal_error"
    UNAUTHORIZED = "unauthorized"
    BAD_REQUEST = "bad_request"
    PERMISSION_DENIED = "permission_denied"
    NOT_IMPLEMENTED = "not_implemented"
    UNEXPECTED = "unexpected_error"
    UNSUPPORTED = "unsupported_operation"


ErrorHandler = (
    Callable[
        [Exception, FunctionType, ErrorResp, str | ErrorCodes | None],
        ErrorResp | Any,
    ]
    | None
)
ErrorMap = List[Tuple[Any, ErrorHandler, str | ErrorCodes | None]]


# Custom error
class ConnectorError(Exception):
    """
    Base exception class for Lumos connectors.
    Preferred way to raise exceptions inside the conenctors.
    `raise ConnectorError(message, error_code)`

    message: str (Custom error message)
    error_code: str | ErrorCodes (The actual error code, eg. "internal_error")
    """

    def __init__(self, message: str, error_code: str | ErrorCodes):
        self.error_code = error_code.value if isinstance(error_code, ErrorCodes) else error_code
        self.message = message


class ExceptionHandler:
    """
    Abstract class for handling exceptions.
    """

    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def handle(
        e: Exception,
        original_func: FunctionType,
        response: ErrorResp,
        error_code: str | ErrorCodes | None = None,
    ) -> ErrorResp:
        """
        Handle an exception. (ErrorHandler signature typing)

        e: Exception (An exception that was raised)
        original_func: FunctionType (The original method that was called, eg. validate_credentials)
        response: ErrorResp (The output of the connector call)
        error_code: str | ErrorCodes | None (The actual error code, eg. "internal_error")
        """
        return response


class DefaultHandler(ExceptionHandler):
    """
    Default exception handler that handles the basic HTTPX/GQL extraction (etc.) and chains onto the global handler.
    """

    @staticmethod
    def handle(
        e: Exception,
        original_func: FunctionType,
        response: ErrorResp,
        error_code: str | ErrorCodes | None = None,
    ) -> ErrorResp:
        status_code: int | None = None

        # HTTPX HTTP Status code
        if hasattr(e, "response") and hasattr(e.response, "status_code"):
            status_code = e.response.status_code
        # GraphQL error code
        if hasattr(e, "code"):
            status_code = e.code

        # Populating some base info
        response.response.message = e.message if hasattr(e, "message") else str(e)
        response.response.status_code = status_code
        response.response.raised_in = f"{original_func.__module__}:{original_func.__name__}"
        response.response.raised_by = f"{e.__class__.__name__}"

        # ConnectorError already has an error code attached, so we need to chain
        if isinstance(e, ConnectorError):
            response.response.error_code = (
                f"{error_code}.{e.error_code}" if error_code else f"{e.error_code}"
            )
        else:
            # Otherwise, it is an unexpected error from an app_id
            response.response.error_code = (
                f"{error_code}.{ErrorCodes.UNEXPECTED.value}"
                if error_code
                else f"{ErrorCodes.UNEXPECTED.value}"
            )

        return response



from connector.async_.lumos import LumosCommandsMixin
from connector.serializers.lumos import (
    AssignEntitlementResp,
    FindEntitlementAssociationsResp,
    FoundAccountData,
    FoundEntitlementData,
    FoundResourceData,
    GetAccountResp,
    GetResourceResp,
    ListAccountsResp,
    ListEntitlementsResp,
    ListResourcesResp,
    UnassignEntitlementResp,
    ValidateCredentialsResp,
    ListCustomAttributesSchemaArgsBase,
    ListCustomAttributesSchemaResp,
)

from {name}.serializers.lumos import (
    AssignEntitlementArgs,
    FindEntitlementAssociationsArgs,
    ListAccountsArgs,
    ListEntitlementsArgs,
    ListResourcesArgs,
    UnassignEntitlementArgs,
    ValidateCredentialsArgs,
)


class AsyncCommands(LumosCommandsMixin):
    app_id = "{hyphenated_name}"

    async def validate_credentials(self, args: ValidateCredentialsArgs) -> ValidateCredentialsResp:
        return ValidateCredentialsResp(
            response=True,
            errors=[],
            raw_data={{}},
        )

    async def list_accounts(self, args: ListAccountsArgs) -> ListAccountsResp:
        return ListAccountsResp(
            response=[FoundAccountData(given_name="{title}", integration_specific_id="100", email="albus@lumos.com")],
            errors=[],
            raw_data = {{"localhost:8080/test": {{"test": "test"}}}},
        )

    async def list_resources(self, args: ListResourcesArgs) -> ListResourcesResp:
        return ListResourcesResp(
            response={{
                "MOCK_CONNECTOR_RESOURCE_1": [
                    FoundResourceData(
                        integration_specific_id="1234",
                        label="Starter Resource",
                        resource_type="MOCK_CONNECTOR_RESOURCE_1",
                    )
                ]
            }},
            errors=[],
            raw_data={{}},
        )

    async def list_entitlements(self, args: ListEntitlementsArgs) -> ListEntitlementsResp:
        return [
            FoundEntitlementData(
                integration_specific_id="1234",
                integration_specific_resource_id="1234",
                is_assignable=False,
                label="mock entitlement",
                entitlement_type="MOCK_CONNECTOR_ENTITLEMENT_1",
            )
        ]

    async def find_entitlement_associations(
        self,
        args: FindEntitlementAssociationsArgs,
    ) -> FindEntitlementAssociationsResp:
        return []

    async def assign_entitlement(
        self,
        args: AssignEntitlementArgs,
    ) -> AssignEntitlementResp:
        pass

    async def unassign_entitlement(
        self,
        args: UnassignEntitlementArgs,
    ) -> UnassignEntitlementResp:
        pass

    async def list_custom_attributes_schema(
        self, args: ListCustomAttributesSchemaArgsBase
    ) -> ListCustomAttributesSchemaResp:
        return []



from connector.cli import run_command

from {name}.async_.lumos import AsyncCommands
from {name}.sync_.lumos import SyncCommands


def main():
    run_command(SyncCommands(), AsyncCommands())


if __name__ == "__main__":
    main()



from argparse import Namespace
from pathlib import Path

TEMPLATE_DIR = Path(__file__).parent / "templates"


def snakecase(name: str) -> str:
    return name.lower().replace(" ", "_").replace("-", "_")


def hyphenate(name: str) -> str:
    return name.lower().replace(" ", "-").replace("_", "-")


def namecase(name: str) -> str:
    return name.replace("_", " ").replace("-", " ").title()


class ScaffoldError(Exception):
    pass


def scaffold(args: Namespace):
    directory = args.directory
    name = args.name

    connector_dir = Path(directory)
    try:
        next(connector_dir.iterdir())
    except (StopIteration, FileNotFoundError):
        pass
    else:
        raise FileExistsError(f"Directory {connector_dir} is not empty")
    connector_dir.mkdir(parents=True, exist_ok=True)

    author = input("Author: ")
    author_email = input("Author email: ")

    # Copy template files to the new connector directory
    output_paths = []
    try:
        for template_file in TEMPLATE_DIR.rglob("*"):
            if template_file.is_file():
                with template_file.open() as f:
                    content = f.read()
                content = content.format(
                    author=author,
                    author_email=author_email,
                    name=snakecase(name),
                    title=namecase(name),
                    hyphenated_name=hyphenate(name),
                )
                # Preserve the relative path of the template file
                relative_path = template_file.relative_to(TEMPLATE_DIR)
                output_path = connector_dir / relative_path
                output_path.parent.mkdir(parents=True, exist_ok=True)
                output_paths.append(output_path)
                with output_path.open("w") as f:
                    f.write(content)
    except Exception as e:
        for output_path in output_paths:
            output_path.unlink()
        raise ScaffoldError(f"Failed to scaffold connector: {e}") from e

    (connector_dir / "connector").rename(connector_dir / snakecase(name))



from abc import ABC, abstractmethod
from contextlib import contextmanager
from types import FunctionType
from typing import Callable, Type
from unittest import IsolatedAsyncioTestCase, TestCase

from pydantic import BaseModel

from connector.async_.abstract import AbstractCommands as AbstractCommandsAsync
from connector.helpers import collect_methods, is_pydantic_model
from connector.sync_.abstract import AbstractCommands as AbstractCommandsSync

_test_docs = """Integration tests are vital to ensure we cover the grounds of the full API.
Define the commands attribute with the commands class to be tested.

Override `discover_data` which fetches the mock response data from the API data folder.
"""
_test_integration_docs = """Tests that the inputs for a command are valid, the response matches the
expected output, and that the response can handle pagination if we detect pagination in the mock
data. The unittest can also be run in live mode where mock data is not used. If pagination is
detected, the test will run the command multiple times to ensure pagination works for the command.
"""


class BaseIntegrationTest(ABC):
    @staticmethod
    def hyphenated(name: str) -> str:
        return "-".join(name.split("_"))

    @abstractmethod
    def discover_data(self, method: Callable):
        """
        Fetch the data from the API data folder. Example implementation:

        ```python
        def discover_data(self, method: callable):
            with open(f"api_data/self.hyphenated({method.__name__}).json") as f:
                return json.load(f)
        ```

        The format of the data should be a list of dictionaries with the following keys
        ```json
        [
            {
                "arguments": {"arg1": "value1", "arg2": "value2"},
                "data": {"key": "value"}
            },
            {
                "arguments": {"arg1": "value1", "arg2": "value2"},
                "data": {"key": "value"}
            },
        ]
        ```

        If the data is paginated in any way, you must provide at least 2 pages of data. This is
        to ensure the integration paginates correctly.
        """
        raise NotImplementedError

    POTENTIAL_PAGE_NAMES = ["page", "cursor", "offset", "limit", "next_cursor", "next_page", "next"]

    def is_paginated(self, data) -> bool:
        """Detect if the data is paginated"""
        if isinstance(data, dict):
            for key in self.POTENTIAL_PAGE_NAMES:
                if key in data:
                    return True
        return False

    @contextmanager
    @abstractmethod
    def mock_method_call(self, method, **kwargs):
        raise NotImplementedError

    def get_args(self, item: dict, method: FunctionType):
        arguments = item.get("arguments")
        response = item.get("data")
        annotations = method.__annotations__

        # Look for the Pydantic model as the args
        for key, value in annotations.items():
            if key == "return":
                continue
            args = [value(**arguments)]
            break
        else:
            args = []
        return arguments, response, args

    def assertion(
        self,
        i: int,
        method: FunctionType,
        data: list,
        response: dict,
        output: BaseModel,
        annotations: dict[str, type],
    ):
        if i == 0:
            assert not (self.is_paginated(response) and len(data) < 2), (
                "Provided data is paginated but only one page is provided. "
                "Please provide at least 2 pages of data."
            )

        if is_pydantic_model(ret := annotations.get("return")):
            if method.__name__ == "info":
                # It's difficult to do a deep comparison of the JSON schemas
                self.assertEqual(output.app_id, response["app_id"])  # type: ignore
                self.assertEqual(output.capabilities, response["capabilities"])  # type: ignore
            else:
                self.assertEqual(output, ret(**response))  # type: ignore
        else:
            self.assertEqual(output, response)  # type: ignore


class SyncIntegrationTest(BaseIntegrationTest, TestCase, ABC):
    sync_commands: Type[AbstractCommandsSync]
    __doc__ = _test_docs

    @classmethod  # type: ignore[misc]
    @property
    def __test__(cls):
        return cls != SyncIntegrationTest

    def setUp(self):
        self.commands = collect_methods(self.sync_commands())

    def test_integration(self):
        for method in self.commands:
            with self.subTest(case=method.__name__):
                data = self.discover_data(method)
                for i, item in enumerate(data):
                    arguments, response, args = self.get_args(item, method)
                    # FIXME For some reason, patching deletes the annotations :skull:
                    annotations = method.__annotations__.copy()
                    with self.mock_method_call(method, args=arguments, data=response):
                        output = method(*args)
                    self.assertion(i, method, data, response, output, annotations)

    test_integration.__doc__ = _test_integration_docs


class AsyncIntegrationTest(BaseIntegrationTest, IsolatedAsyncioTestCase, ABC):
    async_commands: Type[AbstractCommandsAsync]
    __doc__ = _test_docs

    @classmethod  # type: ignore[misc]
    @property
    def __test__(cls):
        return cls != AsyncIntegrationTest

    async def asyncSetUp(self):
        self.commands = collect_methods(self.async_commands())

    async def test_integration(self):
        for method in self.commands:
            with self.subTest(case=method.__name__):
                data = self.discover_data(method)
                for i, item in enumerate(data):
                    arguments, response, args = self.get_args(item, method)
                    annotations = method.__annotations__.copy()
                    with self.mock_method_call(method, args=arguments, data=response):
                        output = await method(*args)
                    self.assertion(i, method, data, response, output, annotations)

    test_integration.__doc__ = _test_integration_docs



from connector.helpers import collect_methods, validate_commands
from connector.serializers.abstract import Info
from connector.utils.info import generate_capability_schema


class AbstractCommands:
    app_id: str

    def __init__(self, app_id: str = ""):
        self.app_id = self.app_id or app_id
        self.validate()

    def validate(self) -> None:
        validate_commands(self)

    def info(self) -> Info:
        if not hasattr(self, "app_id"):
            raise NotImplementedError("app_id attribute is required")
        methods = collect_methods(self)
        capabilities, objects = generate_capability_schema(methods)

        return Info(
            app_id=self.app_id,
            capabilities=sorted(capabilities),
            capability_schema=objects,
        )



from pydantic import BaseModel

__all__ = ("PydanticModel",)


def _set_pydantic_model():
    models = [BaseModel]

    try:
        from pydantic.v1 import BaseModel as BaseModelV1

        models.append(BaseModelV1)
    except ImportError:
        pass

    return frozenset(models)


PydanticModel = _set_pydantic_model()



import inspect
import json
import subprocess
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import cast

from connector.helpers import collect_methods, get_pydantic_model

__all__ = ("collect_commands", "run_command")


# Hacking commands
# ----------------


def _prep_hacking_command(args: Namespace):
    data = vars(args)
    data.pop("command")
    data.pop("func")
    return data


def unasync(
    sync_commands: object, directory: str = "", check: bool = False, quiet: bool = False
) -> None:
    from connector.make_sync import make_sync

    if not directory:
        _dir = Path(inspect.getfile(sync_commands.__class__)).parent.parent / "async_"
    else:
        _dir = Path(directory)

    success = make_sync(Path(_dir), check=check, quiet=quiet)
    if check:
        if not quiet:
            print("Unasync check complete")
        exit(bool(not success))
    else:
        if not quiet:
            print("Unasync complete")


def http_server(async_commands: object, port: int = 8000):
    from connector.http_server import collect_routes, runserver

    router = collect_routes(async_commands)
    try:
        runserver(router, port)
    except KeyboardInterrupt:
        pass


def build_executable(path: str) -> None:
    try:
        subprocess.run(["pyinstaller", "--version"], check=True)
    except FileNotFoundError:
        print("PyInstaller not found in PATH. Please pip install pyinstaller")
        return

    command = [
        "pyinstaller",
        path,
        "--noconsole",
        "--onefile",
        "--clean",
        "--paths=projects/libs/python",
    ]
    if __file__ not in "site-packages":
        command.append("--paths=projects/libs/python")
    subprocess.run(command)


def run_test():
    subprocess.run(["pytest", "tests/"], check=True)


def create_hacking_parser(
    sync_commands: object, async_commands: object, parser: ArgumentParser
) -> None:
    subparsers = parser.add_subparsers(dest="command")

    unasync_parser = subparsers.add_parser("unasync")
    unasync_parser.add_argument(
        "--directory",
        "-d",
        type=str,
        help="The directory to unasync. Defaults to the async_ directory in the connector.",
    )
    unasync_parser.add_argument(
        "--check",
        "-c",
        action="store_true",
        help="Check if the async functions can be converted to sync.",
    )
    unasync_parser.add_argument(
        "--quiet", "-q", action="store_true", help="Do not print anything to the console."
    )
    unasync_parser.set_defaults(
        func=lambda args: unasync(sync_commands, **_prep_hacking_command(args))
    )

    http_server_parser = subparsers.add_parser("http-server")
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_server(async_commands, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    test_parser = subparsers.add_parser("test")
    test_parser.set_defaults(func=lambda args: run_test())

    return None


# Actual Commands
# ---------------


def _print_pydantic(model):
    # Pydantic v2
    if hasattr(model, "model_dump_json"):
        print(model.model_dump_json())
    # Pydantic v1
    elif hasattr(model, "json"):
        print(model.json())
    elif type(model) in (dict, list):
        print(json.dumps(model))
    else:
        print(model)


def _print_decorator(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, list):
            for item in result:
                _print_pydantic(item)
        else:
            _print_pydantic(result)
        return result

    return wrapper


def command_executor(sync_commands: object, args: Namespace):
    """Executes a command from the CLI."""
    method = getattr(sync_commands, cast(str, args.command).replace("-", "_"))
    try:
        model_cls = get_pydantic_model(method.__annotations__)
    except ValueError:
        model_cls = None

    if model_cls:
        try:
            model = model_cls.model_validate_json(args.json)
        except AttributeError:
            model = model_cls.parse_raw(args.json)
        output = method(model)
    else:
        output = method()
    _print_pydantic(output)


def collect_commands(
    sync_commands: object, async_commands: object, no_print: bool = False
) -> ArgumentParser:
    """
    Collect all methods from an object and create a CLI command for each.
    The object must be instantiated.
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    commands = collect_methods(sync_commands)
    for method in commands:
        subparser = subparsers.add_parser(
            method.__name__.replace("_", "-"), description=method.__doc__
        )

        try:
            get_pydantic_model(method.__annotations__)
        except ValueError:
            pass
        else:
            subparser.add_argument("--json", type=str, help="JSON input", required=True)

        subparser.set_defaults(func=lambda args: command_executor(sync_commands, args))

    hacking_subparser = subparsers.add_parser("hacking")
    create_hacking_parser(sync_commands, async_commands, hacking_subparser)

    return parser


def run_command(sync_commands: object, async_commands: object, no_print: bool = False) -> None:
    """Run a command from the CLI."""
    parser = collect_commands(sync_commands, async_commands, no_print)
    args = parser.parse_args()
    args.func(args)



"""
All connector implementations must implement both sync and async versions of the connector.
The easiest way is to create the async version and use "unasync" to generate the sync version.
"""
import filecmp
import tempfile
from pathlib import Path

import unasync

__all__ = ("make_sync", "walk", "Rule")


def walk(path: Path):
    """Recursively walk a directory and yield all files."""
    for p in Path(path).iterdir():
        if p.is_dir():
            yield from walk(p)
            continue
        yield p.resolve()


class Rule(unasync.Rule):
    """Custom unasync rule for easier handling of token replacements"""


def make_sync(
    async_dir: Path,
    rules: list[Rule] | None = None,
    additional_replacements: dict[str, str] | None = None,
    *,
    check: bool = False,
    quiet: bool = False,
) -> bool:
    """
    Generate sync versions of async files.
    :param async_dir: Path to the directory containing the async files.
    :param rules: List of unasync rules.
    :param additional_replacements: Additional tokens to replace.
    :param check: Check if the generated files are different from the existing ones.
    Pass in your usual arguments for rules or leave blank.
    :param quiet: Suppress console output.
    :return: whether operation was successful
    """
    if not rules:
        rules = [unasync.Rule("/async_/", "/sync_/")]
    else:
        assert all(
            isinstance(rule, Rule) for rule in rules
        ), "Rules must be of custom Rule from connector.make_sync."

    if not additional_replacements:
        additional_replacements = {
            # httpx
            "AsyncClient": "Client",
            # imports
            "async_": "sync_",
        }
    for rule in rules:
        rule.token_replacements.update(additional_replacements)

    if check:
        success = True
        for x in walk(async_dir):
            if x.suffix != ".py":
                continue
            async_file = str(x.absolute())
            with tempfile.TemporaryDirectory() as tmp_dir_str:
                temp_dir = Path(tmp_dir_str)
                new_rules = []
                for rule in rules:
                    new_rule = Rule(
                        fromdir=str(Path(async_file).parent.absolute()), todir=str(temp_dir)
                    )
                    new_rule.token_replacements = rule.token_replacements
                    new_rules.append(new_rule)
                unasync.unasync_files([async_file], rules=new_rules)

                for rule in rules:
                    sync_file = Path(async_file.replace(rule.fromdir, rule.todir))
                    if sync_file.exists():
                        break
                else:
                    success = False
                    if not quiet:
                        print(
                            f"Sync file for {Path(async_file).relative_to(async_dir.parent)} does "
                            f"not exist"
                        )
                    continue
                if not filecmp.cmp(temp_dir / x.name, sync_file):
                    success = False
                    if not quiet:
                        print(
                            f"Sync file for {Path(async_file).relative_to(async_dir.parent)} is "
                            f"different"
                        )

        return success

    unasync.unasync_files(
        [str(x.absolute()) for x in walk(async_dir) if x.suffix == ".py"], rules=rules
    )
    return True



from pydantic import Field

from connector.serializers.lumos import (
    AssignEntitlementArgsBase,
    FindEntitlementAssociationsArgsBase,
    GetAccountArgsBase,
    GetResourceArgsBase,
    ListAccountsArgsBase,
    ListEntitlementsArgsBase,
    ListResourcesArgsBase,
    UnassignEntitlementArgsBase,
    ValidateCredentialsArgsBase,
)

from .auth import Auth


class ListAccountsArgs(Auth, ListAccountsArgsBase):
    page: int
    size: int = Field(5)


class ValidateCredentialsArgs(Auth, ValidateCredentialsArgsBase):
    pass


class GetAccountArgs(Auth, GetAccountArgsBase):
    account_id: int


class ListResourcesArgs(Auth, ListResourcesArgsBase):
    page: int
    size: int = Field(5)


class GetResourceArgs(Auth, GetResourceArgsBase):
    pass


class ListEntitlementsArgs(Auth, ListEntitlementsArgsBase):
    page: int
    size: int = Field(5)


class FindEntitlementAssociationsArgs(Auth, FindEntitlementAssociationsArgsBase):
    pass


class AssignEntitlementArgs(Auth, AssignEntitlementArgsBase):
    pass


class UnassignEntitlementArgs(Auth, UnassignEntitlementArgsBase):
    pass



from abc import ABC, abstractmethod

from connector.serializers.lumos import (
    AssignEntitlementArgsBase,
    AssignEntitlementResp,
    FindEntitlementAssociationsArgsBase,
    FindEntitlementAssociationsResp,
    ListAccountsArgsBase,
    ListAccountsResp,
    ListCustomAttributesSchemaArgsBase,
    ListCustomAttributesSchemaResp,
    ListEntitlementsArgsBase,
    ListEntitlementsResp,
    ListResourcesArgsBase,
    ListResourcesResp,
    UnassignEntitlementArgsBase,
    UnassignEntitlementResp,
    ValidateCredentialsArgsBase,
    ValidateCredentialsResp,
)
from connector.sync_.abstract import AbstractCommands

__all__ = ("LumosCommandsMixin",)


class LumosCommandsMixin(AbstractCommands, ABC):
    @abstractmethod
    def list_accounts(self, args: ListAccountsArgsBase) -> ListAccountsResp:
        """
        List a page of accounts.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_credentials(self, args: ValidateCredentialsArgsBase) -> ValidateCredentialsResp:
        """
        Validate that given credentials can be used for base
        integration capabilities.
        """
        raise NotImplementedError

    @abstractmethod
    def list_resources(self, args: ListResourcesArgsBase) -> ListResourcesResp:
        """
        Fetches a list of resources of the given types.
        """
        raise NotImplementedError

    @abstractmethod
    def list_entitlements(
        self,
        args: ListEntitlementsArgsBase,
    ) -> ListEntitlementsResp:
        """
        Fetches a list of entitlements for each of the resources. Both
        directly assignable and unassignable to accounts.
        """
        raise NotImplementedError

    @abstractmethod
    def find_entitlement_associations(
        self, args: FindEntitlementAssociationsArgsBase
    ) -> FindEntitlementAssociationsResp:
        """
        Finds associations between accounts and entitlements,
        which are tied to resources.
        """
        raise NotImplementedError

    @abstractmethod
    def assign_entitlement(self, args: AssignEntitlementArgsBase) -> AssignEntitlementResp:
        """
        Assigns an account directly to an entitlement.

        Requires that the account, resource & entitlement exist
        in the third party application.
        """
        raise NotImplementedError

    @abstractmethod
    def unassign_entitlement(self, args: UnassignEntitlementArgsBase) -> UnassignEntitlementResp:
        """
        Unassigns an account directly from an entitlement.

        Requires that the account, resource & entitlement exist
        in the third party application.
        """
        raise NotImplementedError

    @abstractmethod
    def list_custom_attributes_schema(
        self, args: ListCustomAttributesSchemaArgsBase
    ) -> ListCustomAttributesSchemaResp:
        """
        Find the schema to use for custom attributes.
        """
        raise NotImplementedError



from pydantic import BaseModel, Field, Json


class EmptyModel(BaseModel):
    pass


class CommandTypes(BaseModel):
    argument: Json = Field(..., description="OpenAPI Component")
    output: Json = Field(..., description="OpenAPI Component")


class Info(BaseModel):
    app_id: str
    capabilities: list[str]
    capability_schema: dict[str, CommandTypes]



from enum import Enum
from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, Field


class FoundAccountData(BaseModel):
    integration_specific_id: str
    email: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None
    username: Optional[str] = None
    user_status: Optional[str] = None
    extra_data: Optional[Dict[str, Any]] = None
    custom_attributes: Optional[Dict[str, str]] = None


class CustomAttributeType(str, Enum):
    """Indicates the type of the attribute"""

    STRING = "STRING"
    USER = "USER"


class CustomAttributeCustomizedType(str, Enum):
    """Indicates the type of the entity that would own the attribue"""

    ACCOUNT = "ACCOUNT"
    ENTITLEMENMT = "ENTITLEMENT"
    RESOURCE = "RESOURCE"


class CustomAttributeSchema(BaseModel):
    customized_type: CustomAttributeCustomizedType
    name: str
    attribute_type: CustomAttributeType


class FoundResourceData(BaseModel):
    integration_specific_id: str
    label: str
    resource_type: str
    extra_data: Optional[Dict[str, Any]] = None


class FoundEntitlementData(BaseModel):
    integration_specific_id: str
    integration_specific_resource_id: str
    entitlement_type: str
    is_assignable: Optional[bool] = None
    label: Optional[str] = None
    extra_data: Optional[Dict[str, Any]] = None


class FoundEntitlementAssociation(BaseModel):
    integration_specific_entitlement_id: str
    account: FoundAccountData
    integration_specific_resource_id: str


class EncounteredErrorResponse(BaseModel):
    message: str
    status_code: Optional[int] = None
    error_code: Optional[str] = None
    raised_by: Optional[str] = None
    raised_in: Optional[str] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


Response = TypeVar("Response")


class PageCursor(BaseModel):
    next: str | None


class ResponseWrapper(BaseModel, Generic[Response]):
    response: Response
    raw_data: Optional[dict[str, Any]] = None
    cursor: PageCursor | None = None


class BaseArgs(BaseModel):
    include_raw_data: bool = Field(default=False)


class ListAccountsArgsBase(BaseArgs):
    custom_attributes: Optional[list[str]] = Field(None)


class ListCustomAttributesSchemaArgsBase(BaseArgs):
    pass


class ListCustomAttributesSchemaResp(ResponseWrapper[list[CustomAttributeSchema]]):
    pass


class ListAccountsResp(ResponseWrapper[list[FoundAccountData]]):
    pass


class ValidateCredentialsArgsBase(BaseArgs):
    pass


class ValidateCredentialsResp(ResponseWrapper[bool]):
    pass


class GetAccountArgsBase(BaseArgs):
    pass


class GetAccountResp(ResponseWrapper[FoundAccountData]):
    pass


class ListResourcesArgsBase(BaseArgs):
    resource_type: str


class ListResourcesResp(ResponseWrapper[list[FoundResourceData]]):
    pass


class GetResourceArgsBase(BaseArgs):
    resource_type: str
    integration_specific_id: str


class GetResourceResp(ResponseWrapper[FoundResourceData]):
    pass


class ListEntitlementsArgsBase(BaseArgs):
    resource_type: str
    resource_integration_specific_id: str


class ListEntitlementsResp(ResponseWrapper[list[FoundEntitlementData]]):
    pass


class FindEntitlementAssociationsArgsBase(BaseArgs):
    entitlements: List[FoundEntitlementData]


class FindEntitlementAssociationsResp(ResponseWrapper[list[FoundEntitlementAssociation]]):
    pass


class AssignEntitlementArgsBase(BaseArgs):
    account: FoundAccountData
    entitlement: FoundEntitlementData


class AssignEntitlementResp(ResponseWrapper[bool]):
    pass


class UnassignEntitlementArgsBase(BaseArgs):
    account: FoundAccountData
    entitlement: FoundEntitlementData


class UnassignEntitlementResp(ResponseWrapper[bool]):
    pass


class ErrorResp(ResponseWrapper[EncounteredErrorResponse]):
    error: bool = True
    pass
